/*
 ============================================================================
 Name        : CUDA_MatrixMul.cu
 Author      : 
 Version     :
 Copyright   : 
 Description : CUDA compute reciprocals
 ============================================================================
 */

#include <iostream>
#include <numeric>
#include <stdlib.h>

static void CheckCudaErrorAux(const char *, unsigned, const char *,
		cudaError_t);
#define CUDA_CHECK_RETURN(value) CheckCudaErrorAux(__FILE__,__LINE__, #value, value)

/**
 * Check the return value of the CUDA runtime API call and exit
 * the application if the call has failed.
 */
static void CheckCudaErrorAux(const char *file, unsigned line,
		const char *statement, cudaError_t err) {
	if (err == cudaSuccess)
		return;
	std::cerr << statement << " returned " << cudaGetErrorString(err) << "("
			<< err << ") at " << file << ":" << line << std::endl;
	exit(1);
}

float *generate_data(int length) {
	float *data = (float *) malloc(sizeof(float) * length);
	for (int i = 0; i < length; i++) {
		data[i] = ((float) (rand() % 20) - 5) / 5.0f;
	}
	return data;
}

float *generate_fixed_data(int length) {
	float *data = (float *) malloc(sizeof(float) * length);
	int value = 1;
	for (int i = 0; i < length; i++) {
		data[i] = value++;
	}
	return data;
}

// Compute C = A * B
__global__ void matrixMultiply(float *A, float *B, float *C, int numARows,
		int numAColumns, int numBRows, int numBColumns) {
	int row = blockIdx.y * blockDim.y + threadIdx.y;
	int col = blockIdx.x * blockDim.x + threadIdx.x;
	if (row < numARows && col < numBColumns) {
		float sum = 0;
		for (int ii = 0; ii < numAColumns; ii++) {
			sum += A[row * numAColumns + ii] * B[ii * numBColumns + col];
		}
		C[row * numBColumns + col] = sum;
	}
}

void print_matrix(float* matrix, int rows, int cols) {
	for (int i = 0; i < rows; i++) {
		for (int j = 0; j < cols; j++) {
			printf("%f ", matrix[i * cols + j]);
		}
		printf("\n");
	}
	printf("\n");
}

int main() {
	float *hostA; // The A matrix
	float *hostB; // The B matrix
	float *hostC; // The output C matrix
	float *deviceA;
	float *deviceB;
	float *deviceC;
	int numARows = 2;    // number of rows in the matrix A
	int numAColumns = 3; // number of columns in the matrix A
	int numBRows = 3;    // number of rows in the matrix B
	int numBColumns = 4; // number of columns in the matrix B
	int numCRows = numARows;
	int numCColumns = numBColumns;
	// allocate host memory
	hostA = generate_fixed_data(numARows * numAColumns);
	hostB = generate_fixed_data(numBRows * numBColumns);
	hostC = (float*) malloc(numCRows * numCColumns * sizeof(float));
	// allocate device memory
	CUDA_CHECK_RETURN(
			cudaMalloc((void ** )&deviceA,
					sizeof(float) * numARows * numAColumns));
	CUDA_CHECK_RETURN(
			cudaMalloc((void ** )&deviceB,
					sizeof(float) * numBRows * numBColumns));
	CUDA_CHECK_RETURN(
			cudaMalloc((void ** )&deviceC,
					sizeof(float) * numCRows * numCColumns));
	// copy from host to device memory
	CUDA_CHECK_RETURN(
			cudaMemcpy(deviceA, hostA, numARows * numAColumns * sizeof(float),
					cudaMemcpyHostToDevice));
	CUDA_CHECK_RETURN(
			cudaMemcpy(deviceB, hostB, numBRows * numBColumns * sizeof(float),
					cudaMemcpyHostToDevice));
	// organize grid
	dim3 blockDim(16, 16);
	dim3 gridDim(ceil(((float) numAColumns) / blockDim.x),
			ceil(((float) numBRows) / blockDim.y));
	// execute kernel
	matrixMultiply<<<gridDim, blockDim>>>(deviceA, deviceB, deviceC, numARows,
			numAColumns, numBRows, numBColumns);
	cudaDeviceSynchronize();
	// copy results from device memory to host
	CUDA_CHECK_RETURN(
			cudaMemcpy(hostC, deviceC, numARows * numBColumns * sizeof(float),
					cudaMemcpyDeviceToHost));
	print_matrix(hostA, numARows, numAColumns);
	print_matrix(hostB, numBRows, numBColumns);
	print_matrix(hostC, numCRows, numCColumns);
	// Free the GPU memory here
	cudaFree(deviceA);
	cudaFree(deviceB);
	cudaFree(deviceC);
	// Free host memory
	free(hostA);
	free(hostB);
	free(hostC);
}
